\iilo.xpl	2-Mar-2018
\Optimizing code generator.
\This reads in the .iil file produced by xpl0 and outputs an ARM .s file.
\To understand this code you should first learn how iil.xpl works.
\
\This program is free software; you can redistribute it and/or modify it under
\ the terms of the GNU General Public License version 2 as published by the
\ Free Software Foundation.
\This program is distributed in the hope that it will be useful, but WITHOUT
\ ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
\ FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
\ details.
\You should have received a copy of the GNU General Public License along with
\ this program (in the file LICENSE.TXT); if not, write to the Free Software
\ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\
\The authors can be reached at: loren.blaney@gmail.com
\
\REVISIONS:
\  4-Oct-2015: Version 0.9 released.
\ 17-Dec-2015: V0.92, Changed IMM labels so they work with linked library.
\  Fixed bug in GenADROff where an existing variable in r0 was not bombed.
\  Increase VStkSize from 100 to 10000, so many global arrays can be declared.
\ 12-Aug-2016: V1.0 release.
\ 29-May-2017: V2.0, Bomb register variables following in-line assembly code.
\  Don't use "/" for command-line switch because it prevents use of path names.
\ 30-Jun-2017: V2.1, Fix bug when optimized proc is nested two or more levels.
\ 02-Mar-2018: Version 2.2, add intrinsics Floor, Ceil, and Pow to GenCML.
\
\Register usage:
\ r0:	 xpl0's top of stack (TOS)
\ r1-r4: copies of memory variables and constants
\ r5-r6: declared global arrays
\ r7:	 'for' loop control variable
\ r8:	 access to runtime code (natrpi.s) variables
\ r9:	 heap pointer
\ r10:	 access to global (heap) variables
\ r11:	 access to local (heap) variables
\ r12:	 access to intermediate level (i.e. other heap) variables
\ r13:	 hardware stack pointer (sp)
\ r14:	 link register (lr): subroutine calls and scratch
\ r15:	 program counter (pc)

string	0;		\null-terminated strings used throughout

int	OldLev,		\level accessing non-local or non-global variables (r12)
	ErrCnt,		\error counter
	OD,		\output device number (3 or 7) for assembly code
	LastOp;		\immediately preceding I2L (aka: iil) opcode

def	EOF=$1A, Bel=$07, LF=$0A, CR=$0D;
def	IntSize=4;	\number of bytes in an integer

def	VStkSize=10000,	\maximum number of items in virtual stack
	LevelSize=8;	\maximum number of static nesting levels (0 thru 7)

int	VSP,		\Virtual Stack Pointer for these stacks (push = sto inc)
	VSTyp,		\virtual stack: holds argument type Reg Stk Adr Imm Real
	VSLev,		\virtual stack: holds level of adr argument or imm value
	VSOff;		\virtual stack: holds offset of adr argument or ^C or ^L
def	\types of virtual-stack arguments:
	RegTyp,		\TOS (top of virtual stack) is in register r0
	StkTyp,		\TOS is on the hardware stack
	AdrTyp,		\TOS is in memory at an address (level, offset)
	ImmTyp,		\TOS is immediate value (^C=normal constant; ^L=label)
	RealTyp;	\TOS is a floating point value on the hardware stack

int	RegLev(16),	\register N contains copy of variable at Lev and Off
	RegOff(16),
	RegImm(16),	\flag: register N contains immediate constant in RegVal
	RegVal(16),	\constant value contained in register N

	RegMax,		\maximum register number used for copies of variables
	RegLastUsed,	\used by func GetReg to get next register [1..RegMax]
	NeedBase,	\a set bit corresponds to level needing to set up base
			\ because r12 accessed an intermediate-level variable
	ProcRets,	\array: code labels for procedure returns for each level
	Level,		\static nesting level of current procedure
	LineCnt;	\line counter for error messages
int	FSP;		\FPU Stack Pointer for "d" registers (push = sto inc)
def	ForStkSize=100;
int	ForSP,
	ForStkLev(ForStkSize),
	ForStkOff(ForStkSize);



proc	Error(N);		\Display an error message
int	N;
int	Err, Ch, I;
char	Str;
def	MaxErr=3;
[Err:= Reserve((MaxErr+1)*IntSize);
for I:= 0 to MaxErr do Err(I):= "? ";	\unused errors ="?"

Err(1):= "Out of range";
Err(2):= "Unrecognized iil code";
Err(3):= "L or C expected";

ChOut(0, Bel);
CrLf(0);
Text(0, "***** Error no. ");  IntOut(0, N);
Text(0, " on line ");  IntOut(0, LineCnt);  Text(0, " *****");
CrLf(0);
if N<0 ! N>MaxErr then N:= 0;
Str:= Err(N);
I:= 0;
loop	[Ch:= Str(I);			\output message
	if Ch = 0 then quit;

	if Ch=^* then Text(0, " EXPECTED BUT NOT FOUND")
		else ChOut(0, Ch);
	I:= I+1];
CrLf(0);
Text(0, "Attempt to continue (y/n)? ");
OpenI(0);
case ChIn(0) of ^N, ^n: exit 1 other [];
OD:= 7;					\discard the output file
ErrCnt:= ErrCnt+1;
];	\Error



proc	Txt(Str, R0, R1, R2);	\Output string of assembly code
char	Str;
int	R0, R1, R2;		\need be passed if %, $, and ? are not used
int	I, Ch;
[for I:= 0 to 32000 do
	[Ch:= Str(I);
	case Ch of
	  ^%:	IntOut(OD, R0);
	  ^$:	IntOut(OD, R1);
	  ^?:	IntOut(OD, R2);
	  ^|:	CrLf(OD);
	  0:	return
	other	ChOut(OD, Ch);
	];
];	\Txt



proc	BugMsg(S);	\Display message for internal compiler error (i.e: bug)
int	S;
[Text(0, "INTERNAL ERROR - ");  Text(0, S);  CrLf(0);
\Text(0, "Line ");  IntOut(0, LineCnt);  CrLf(0);
Txt("@ INTERNAL ERROR - ");  Txt(S);
Txt(" on line ");  IntOut(OD, LineCnt);  CrLf(OD);
Error(0);
];	\BugMsg



proc	HexB(B);		\Output a byte in hex format, e.g: 4F
int	B;
char	HexDigit;
[HexDigit:= "0123456789ABCDEF ";
ChOut(OD, HexDigit(B>>4 & $0F));
ChOut(OD, HexDigit(B & $0F));
];	\HexB



proc	NewLevel(Lev);\Set up r12 to access intermediate variables, if necessary
int	Lev;
[if Lev#0 & Lev#Level & Lev#OldLev then
	[Txt("	ldr	r12, [r8, #base");  IntOut(OD, Lev); Txt("-dseg]|");
	OldLev:= Lev;
	NeedBase:= NeedBase ! 1<<Lev;	\mark level that will need its base to
	];				\ be set up at the start of the body of
];	\NewLevel			   the enclosing procedure



func	RegVar(Lev, Off);
\Return register number containing copy of variable at specified level & offset
int	Lev, Off;
int	N;
[for N:= 0 to 7 do
	if RegLev(N)=Lev & RegOff(N)=Off then return N;
return -1;		\no register contains the specified variable
];	\RegVar



func	RegCon(Con, Typ);	\Return register number containing specified constant
int	Con, Typ;
int	I;
[if Typ # ^C then return -1;
for I:= 0 to 7 do
	if RegVal(I)=Con & RegImm(I) then return I;
return -1;			\no register contains the constant
];	\RegCon



func	VStkIsReg(Depth, Reg);	\Return 'true' if VStack address item is in Reg
int	Depth, Reg;
return VSLev(VSP-Depth)=RegLev(Reg) & VSOff(VSP-Depth)=RegOff(Reg);



func	Rol(N, C);		\Rotate N left C bits
int	N, C;
[C:= $1F & C;			\CPU's instruction limits shift to 0..31 bits
return N<<C ! N>>(IntSize*8-C);
];	\Rol



func	ImmOk(Imm);		\Returns 'true' if the immediate value can be
int	Imm;			\ used in an instruction, e.g: mov r0, #0x128
int	I, Mask;
[Mask:= $FF;
for I:= 0 to 15 do
	[if (Imm & Mask) = Imm then return true;
	Mask:= Rol(Mask, 2);
	];
return false;
];	\ImmOk



proc	MovImm(Reg, Val, Type);	\Generate code to put immediate value into Reg
int	Reg, Val, Type;
int	Ones, Shift, SVal;
[if Type = ^L then	\generate address to a label used for strings and
	[		\ for constant arrays
	Txt("	ldr	r%, KP$  		@0B IMM|", Reg, Val);
	Txt("	b	KK$|", Reg, Val);
	Txt("KP$:	.word	L$|", Reg, Val);
	Txt("KK$:|", Reg, Val);
	]
else if Val=RegVal(Reg) & RegImm(Reg) then []	\Val is already in reg
else if ImmOk(Val) then				\handle 0x8000_0001, etc.
    Txt("	mov	r%, #$  		@0B IMM|", Reg, Val)
else if ImmOk(~Val) then			\handle 0x7FFF_FFFE, etc.
    Txt("	mvn	r%, #$  		@0B IMM|", Reg, ~Val)
else	[SVal:= Val;
	Ones:= 0;		\count number of set (one) bits
	for Shift:= 0 to 31 do
		if 1<<Shift & Val then Ones:= Ones+1;

	if Ones > 16 then Val:= ~Val;
	Shift:= 0;
	if Val then
		while (Val>>Shift & $03) = 0 do Shift:= Shift+2;
	if Ones > 16 then
		Txt("	mvn	r%, #0x", Reg)
	else	Txt("	mov	r%, #0x", Reg);
	HexOut(OD, Val & $FF<<Shift);
	Txt("		@0B IMM|");

	loop	[Shift:= Shift+8;
		if Shift >=32 then quit;
		Val:= Val & -1<<Shift;
		if Val = 0 then quit;
		while (Val>>Shift & $03) = 0 do Shift:= Shift+2;
		if Ones > 16 then
			Txt("	bic	r%, #0x", Reg)
		else	Txt("	orr	r%, #0x", Reg);
		HexOut(OD, Val & $FF<<Shift);
		CrLf(OD);
		];
	Val:= SVal;
	];

RegImm(Reg):= Type # ^L;	\don't use label for immediate value
if RegImm(Reg) then RegVal(Reg):= Val;
RegLev(Reg):= -1;
];	\MovImm



proc	OpString(Str, Typ, Lev, Off, Reg, Reg2);
\Output opcode string with embedded operands or registers.
\ Special string symbols: |=crlf, &=operand, %=register number.
\*** WARNING***: Be sure to bomb r0-RegMax and r7 (RegLev(7):= -1) before using
\ this procedure to store into any memory variable (otherwise it will store into
\ the shadowing register instead).
char	Str;	\string
int	Typ,	\type of operand (&: ^R, ^A or ^I)
	Lev,	\level for address (A), or value for immediate (I)
	Off,	\offset for address, or string for reg (R)
	Reg,	\register number to replace "%"
	Reg2;	\register number to replace "$" (an optional argument)
int	I, Ch, R;
int	J;
char	Str2;
[if Typ=^A then	\if address operand, make sure r12 is properly set if necessary
	if RegVar(Lev, Off) < 0 then NewLevel(Lev);

for I:= 0 to 32000 do
	[Ch:= Str(I);
	case Ch of
	  ^&:	case Typ of
		  ^R:	[Str2:= Off;				\register string
			loop for J:= 0 to 32000 do
				case Str2(J) of
				  ^%:	IntOut(OD, Reg);
				  ^$:	IntOut(OD, Reg2);
				  0:	quit
				other	ChOut(OD, Str2(J));
			];
		  ^A:	[R:= RegVar(Lev, Off);			\register var.
			if R>=0 then [ChOut(OD, ^r);  IntOut(OD, R)]
			else	[case Lev of			\address var.
				  0:	Txt("[r10, #");
				  Level:Txt("[r11, #")
				other	Txt("[r12, #");
				IntOut(OD, Off);  Txt("]");
				];
			];
		  ^I:	[if Off=^L then				\immediate value
				BugMsg("OpString: Off=^^L")
			else [Txt("r%, #", Reg);  IntOut(OD, Lev);
			     if not ImmOk(Lev) then
				BugMsg("OpString: Lev out of range");
			     ];			
			]
		other	[];

	  ^?:	if RegVar(Lev, Off) >= 0 then
			Txt("mov")				\register var.
		else	Txt("ldr");

	  ^%:	IntOut(OD, Reg);

	  ^|:	CrLf(OD);
	  0:	return
	other	ChOut(OD, Ch);
	]
];	\OpString



func	GetReg(Used);			\Return the number of a register to use for TOS
int	Used;		\don't use GetR to assing a register that's already used
int	Reg;

	func	GetR;		\Return next register number [1..RegMax]
	\There are some subtitles here. Don't get too clever and try to select
	\ an unused register. The instruction scheduler works best for
	\ sequentially ascending register numbers. The STX2 I2L instruction
	\ requires that three different registers are used (unless same value).
	[repeat	RegLastUsed:= RegLastUsed+1;
		if RegLastUsed > \RegMax\ 4 then RegLastUsed:= 1;
\debug: use r4 instead of RegMax because r5 and r6 can be global and never get changed
	until	(1<<RegLastUsed & Used) = 0;
	return RegLastUsed;
	];	\GetR

[case VSTyp(VSP-1) of
  RegTyp:			\if TOS is already in r0 then use r0
  	Reg:= 0;
  StkTyp:
	Reg:= GetR;
  AdrTyp:		\if variable is already in a register then use that reg
	[Reg:= RegVar(VSLev(VSP-1), VSOff(VSP-1));
	if Reg < 0 then		\variable is not in a register
		Reg:= GetR;	\ so select a register to load the variable
	];	
  ImmTyp:		\if constant is already in a register then use that reg
	[Reg:= RegCon(VSLev(VSP-1), VSOff(VSP-1));
	if Reg < 0 then Reg:= GetR;
	]
other	BugMsg("GetReg");
return	Reg;
];	\GetReg



proc	PushV(Typ, Lev, Off);	\Push an item onto the virtual stack
int	Typ, Lev, Off;
[if VSP < VStkSize then
	[VSTyp(VSP):= Typ;
	 VSLev(VSP):= Lev;
	 VSOff(VSP):= Off;
	 VSP:= VSP+1;
	]
else BugMsg("PushV overflow");
];	\PushV



proc	DropV(N);		\Drop (pop & discard) N items from virtual stack
int	N;
[VSP:= VSP-N;
if VSP < 0 then [VSP:= 0;  BugMsg("DropV VSP<0")];
];	\DropV



proc	PushFor(Lev, Off);	\Push a variable onto the 'for' loop stack
int	Lev, Off;
[if ForSP < ForStkSize then
	[ForStkLev(ForSP):= Lev;
	 ForStkOff(ForSP):= Off;
	 ForSP:= ForSP+1;
	]
else BugMsg("PushFor overflow");
];	\PushFor



proc	DropFor;		\Pop & discard one item from 'for' loop stack
[ForSP:= ForSP-1;
if ForSP < 0 then [ForSP:= 0;  BugMsg("DropFor ForSP<0")];
];	\DropFor



func	IsZero(Comm);		\Eliminate operations that result in zero
				\Returns 'true' if operation equals 0
int	Comm;	\operation is commutative
[if VSLev(VSP-2)=0 & VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C then
	[DropV(1);  return true];	\e.g: 0/A; drop A leave 0

if Comm then				\e.g: 0*A = A*0 = 0
	[if VSLev(VSP-1)=0 & VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C then
		[VSTyp(VSP-2):= ImmTyp;	\e.g: A*0
		VSLev(VSP-2):= 0;
		VSOff(VSP-2):= ^C;
		DropV(1);		\drop A; leave 0
		return true;
		];
	];
return false;
];	\IsZero



func	Identity(N, Comm);	\Eliminate identity operations like A+0 and A*1
int	N,			\returns 'true' if it was an identity operation
	Comm;	\operation is commutative
[if VSLev(VSP-1)=N & VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C then
	[DropV(1);  return true];

if Comm then
	[if VSLev(VSP-2)=N & VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C then
		[\eliminate identity operator
		VSTyp(VSP-2):= VSTyp(VSP-1);
		VSLev(VSP-2):= VSLev(VSP-1);
		VSOff(VSP-2):= VSOff(VSP-1);
		DropV(1);
		return true;
		];
	];
return false;
];	\Identity



proc	TOS2Reg(R);		\Copy TOS into register specified by 'R'
int	R;			\ (unless it's a real)
[case VSTyp(VSP-1) of
  RegTyp:
	[if R # 0 then
		[OpString("	mov	r%, r0|", 0, 0, 0, R);
		RegLev(R):= RegLev(0);
		RegOff(R):= RegOff(0);
		RegImm(R):= RegImm(0);
		RegVal(R):= RegVal(0);
		];
	];
  StkTyp:
	[OpString("	pop	{r%}|", 0, 0, 0, R);
	RegLev(R):= -1;  RegImm(R):= false;
	];
  AdrTyp:
	if VSLev(VSP-1)#RegLev(R) ! VSOff(VSP-1)#RegOff(R) then
	\if TOS is not already in the register then gen code to put it in
	    [OpString("	?	r%, &|", ^A, VSLev(VSP-1), VSOff(VSP-1), R);
	    RegLev(R):= VSLev(VSP-1);  RegOff(R):= VSOff(VSP-1);
	    RegImm(R):= false;
	    ];
  ImmTyp:
	MovImm(R, VSLev(VSP-1), VSOff(VSP-1))
\ RealTyp: []
other	BugMsg("TOS2Reg");
];	\TOS2Reg



proc	CleanReg(Lev);	\Push any r0 below Lev in virtual stack onto h/w stack
int	Lev;	\ i.e: clean r0 off virtual stack because r0 or the h/w stack
int	I;	\ is about to be used.
[for I:= 0 to VSP-(Lev+1) do
	if VSTyp(I) = RegTyp then
		[Txt("	push	{r0}|");
		VSTyp(I):= StkTyp];
];	\CleanReg



proc	TOS2Stack;		\Put TOS on the h/w stack, if not already there
int	R;
[if VSP < 1 then [BugMsg("TOS2Stack");  return];
CleanReg(1);
case VSTyp(VSP-1) of
  RegTyp:
  	Txt("	push	{r0}|");
  StkTyp:
  	[];
  AdrTyp:
	[R:= RegVar(VSLev(VSP-1), VSOff(VSP-1));
	if R < 0 then		\variable is not in a register, so load it
		[R:= GetReg(0);
		OpString("	ldr	r%, &|", ^A, VSLev(VSP-1), VSOff(VSP-1), R);
		RegLev(R):= VSLev(VSP-1);  RegOff(R):= VSOff(VSP-1);
		RegImm(R):= false;
		];
	Txt("	push	{r%}|", R);
	];	
  ImmTyp:
	[R:= GetReg(0);
	MovImm(R, VSLev(VSP-1), VSOff(VSP-1));
	Txt("	push	{r%}|", R);
	];
  RealTyp:
	return			\it's already on the FPU stack
other	BugMsg("TOS2Stack");
VSTyp(VSP-1):= StkTyp;
];	\TOS2Stack



proc	TOS2R0;			\Put TOS into r0, if it's not already there
[CleanReg(1);
TOS2Reg(0);
VSTyp(VSP-1):= RegTyp;
];	\TOS2R0

\-------------------------------------------------------------------------------

proc	GenOp2(OpStr, Rev, NoImm, Op);	\Generate instructions for 2-operand ops
\e.g:	GenOp2("	add	r0, &|", false, false, "+");
char	OpStr;	\opcode string (&=operand)
int	Rev,	\flag: reverse order of operands (cuz they're not commutative)
	NoImm;	\flag: cannot be done using immediate operand (like mul)
char	Op;	\string containing operator, e.g: ">="
int	NOS, TOS; \operand types: REG, STK, ADR, IMM, (REAL)


\Format used in the following procedure names is: DO_NOS_TOS

proc	DO_REG_REG;		\Can occur because of register variable	 0
[OpString(OpStr, ^R, 0, "r0, r0", 0);				\add r0, r0, r0
];	\DO_REG_REG



proc	DO_STK_REG;						\1
[Txt("	pop	{r14}|");					\nos
OpString(OpStr, ^R, 0, "r14, r0", 0);				\sub r0, r14, r0
];	\DO_STK_REG						 sub r0, nos tos



proc	DO_ADR_REG;						\2
int	R;
[VSP:= VSP-1;
R:= GetReg(0);
TOS2Reg(R);
OpString(OpStr, ^R, 0, "r%, r0", R);			\sub r0, r3, r0
VSP:= VSP+1;
];	\DO_ADR_REG



proc	DO_IMM_REG;						\3
int	R;
[
VSP:= VSP-1;
if ImmOk(VSLev(VSP-1)) & VSOff(VSP-1)=^C & ~NoImm & ~Rev then
	OpString(OpStr, ^I, VSLev(VSP-1), VSOff(VSP-1), 0)	\add r0, r0, #12
else	[R:= GetReg(0);
	MovImm(R, VSLev(VSP-1), VSOff(VSP-1));			\nos
	OpString(OpStr, ^R, 0, "r%, r0", R);			\sub r0, r3, r0
	];							\sub r0, nos tos
VSP:= VSP+1;
];	\DO_IMM_REG



proc	DO_STK_STK;						\4
[Txt("	pop	{r0}|");					\tos
RegLev(0):= -1;  RegImm(0):= false;
Txt("	pop	{r14}|");					\nos
OpString(OpStr, ^R, 0, "r14, r0", 0);				\sub r0, r14, r0
];	\DO_STK_STK						 sub r0, nos tos



proc	DO_ADR_STK;						\5
int	R;
[Txt("	pop	{r14}|");					\pop tos into lr
VSP:= VSP-1;
R:= GetReg(0);
TOS2Reg(R);
OpString(OpStr, ^R, 0, "r%, r14", R);			\sub r0, r3, r14
VSP:= VSP+1;
];	\DO_ADR_STK



proc	DO_IMM_STK;						\6
[Txt("	pop	{r0}|");					\tos
RegLev(0):= -1;  RegImm(0):= false;
DO_IMM_REG;	\3\						\sub r0, nos tos
];	\DO_IMM_STK



proc	DO_REG_ADR;						\7
int	R;
[R:= GetReg(0);
TOS2Reg(R);
OpString(OpStr, ^R, 0, "r0, r%", R)				\sub r0, r0, r7
];	\DO_REG_ADR



proc	DO_STK_ADR;						\8
int	R;
[Txt("	pop	{r14}|");					\nos
R:= GetReg(0);
TOS2Reg(R);
OpString(OpStr, ^R, 0, "r14, r%", R)				\sub r0, r14, r7
];	\DO_STK_ADR



proc	DO_ADR_ADR;						\9
int	Rn, Rt;
[Rt:= GetReg(0);
TOS2Reg(Rt);
VSP:= VSP-1;
Rn:= GetReg(1<<Rt);
TOS2Reg(Rn);
OpString(OpStr, ^R, 0, "r%, r$", Rn, Rt);			\sub r0, r3, r7
VSP:= VSP+1;
];	\DO_ADR_ADR



proc	DO_IMM_ADR;						\10
[OpString("	?	r0, &|", ^A, VSLev(VSP-1), VSOff(VSP-1), 0); \tos
RegLev(0):= VSLev(VSP-1);  RegOff(0):= VSOff(VSP-1);  RegImm(0):= false;
DO_IMM_REG;	\3\						\sub r0, nos tos
];	\DO_IMM_ADR



proc	DO_REG_IMM;						\11
int	R;
[if ImmOk(VSLev(VSP-1)) & VSOff(VSP-1)=^C & ~NoImm then
	OpString(OpStr, ^I, VSLev(VSP-1), VSOff(VSP-1), 0)	\sub r0, r0, #4
else	[\if constant is already in a register then use the register
	R:= RegCon(VSLev(VSP-1), VSOff(VSP-1));
	if R <= 0 then
		[R:= GetReg(0);	\else load constant into a register
		MovImm(R, VSLev(VSP-1), VSOff(VSP-1));
		];
	OpString(OpStr, ^R, 0, "r0, r%", R);			\sub r0, r0, r3
	];							\sub r0, nos tos
];	\DO_REG_IMM



proc	DO_STK_IMM;						\12
[Txt("	pop	{r0}|");					\nos
RegLev(0):= -1;  RegImm(0):= false;
DO_REG_IMM;	\11\						\sub r0, nos tos
];	\DO_STK_IMM



proc	DO_ADR_IMM;						\13
int	Rn, Rt;
[
VSP:= VSP-1;
Rn:= RegVar(VSLev(VSP-1), VSOff(VSP-1));
if Rn < 0 then
	[Rn:= GetReg(0);	\nos is not already in a register, so load one
	OpString("	ldr	r%, &|", ^A, VSLev(VSP-1), VSOff(VSP-1), Rn);
	RegLev(Rn):= VSLev(VSP-1);  RegOff(Rn):= VSOff(VSP-1);
	RegImm(Rn):= false;
	];

VSP:= VSP+1;
if ImmOk(VSLev(VSP-1)) & VSOff(VSP-1)=^C & ~NoImm then
	OpString(OpStr, ^I, VSLev(VSP-1), VSOff(VSP-1), Rn)	\sub r0, r2, #4
else	[Rt:= RegCon(VSLev(VSP-1), VSOff(VSP-1));
	if Rt < 0 then
		[Rt:= GetReg(1<<Rn);
		\tos is not already in a register, so load one
		MovImm(Rt, VSLev(VSP-1), VSOff(VSP-1));
		];
	OpString(OpStr, ^R, 0, "r%, r$", Rn, Rt);		\sub r0, r2, r3
	];							\sub r0, nos tos
];	\DO_ADR_IMM



proc	DO_IMM_IMM;						\14
\WARNING: this must not alter r0 (not "clean")
int	TOSVal, NOSVal, C1, C2, Remain;
[if VSOff(VSP-1)#^C ! VSOff(VSP-2)#^C then BugMsg("DO_IMM_IMM labels");

TOSVal:= VSLev(VSP-1);
NOSVal:= VSLev(VSP-2);
C1:= Op(0);  C2:= Op(1);		\get operator symbol from string
case C1 of
  ^=:	TOSVal:= NOSVal=TOSVal;
  ^#:	TOSVal:= NOSVal#TOSVal;
  ^>:	case C2 of
	  ^=:	TOSVal:= NOSVal>=TOSVal;
	  ^>:	TOSVal:= NOSVal>>TOSVal
	other	TOSVal:= NOSVal>TOSVal;
  ^<:	case C2 of
	  ^=:	TOSVal:= NOSVal<=TOSVal;
	  ^<:	TOSVal:= NOSVal<<TOSVal
	other	TOSVal:= NOSVal<TOSVal;
  ^!:	TOSVal:= NOSVal!TOSVal;
  ^&:	TOSVal:= NOSVal&TOSVal;
  ^|:	TOSVal:= NOSVal|TOSVal;
  ^+:	TOSVal:= NOSVal+TOSVal;
  ^-:	if C2=^> then TOSVal:= NOSVal->>TOSVal
		 else TOSVal:= NOSVal-TOSVal;
  ^*:	TOSVal:= NOSVal*TOSVal;
  ^/:	[if TOSVal=0 then [Error(73);  TOSVal:= $7FFFFFFF;  Remain:= 0]
	else [TOSVal:= NOSVal/TOSVal;  Remain:= Rem(0)];
	MovImm(3, Remain, ^C);
	Txt("	str	r3, [r8, #remain-dseg]|");
	]
other	BugMsg("DO_IMM_IMM");

DropV(2);
PushV(ImmTyp, TOSVal, ^C);
];	\DO_IMM_IMM



[	\GenOp2
if VSP < 2 then BugMsg("GenOp2");

if VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)#^C !	\handle "NAME"+2, etc.
   VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)#^C then TOS2R0;

TOS:= VSTyp(VSP-1);
NOS:= VSTyp(VSP-2);
if TOS#ImmTyp ! NOS#ImmTyp then CleanReg(2);	\if they're not both immediates

\Generate opcode based on NOS and TOS
case of
  NOS=RegTyp & TOS=RegTyp: DO_REG_REG;		\0
  NOS=StkTyp & TOS=RegTyp: DO_STK_REG;		\1
  NOS=AdrTyp & TOS=RegTyp: DO_ADR_REG;		\2
  NOS=ImmTyp & TOS=RegTyp: DO_IMM_REG;		\3

 \NOS=RegTyp & TOS=StkTyp: never occurs
  NOS=StkTyp & TOS=StkTyp: DO_STK_STK;		\4
  NOS=AdrTyp & TOS=StkTyp: DO_ADR_STK;		\5
  NOS=ImmTyp & TOS=StkTyp: DO_IMM_STK;		\6

  NOS=RegTyp & TOS=AdrTyp: DO_REG_ADR;		\7
  NOS=StkTyp & TOS=AdrTyp: DO_STK_ADR;		\8
  NOS=AdrTyp & TOS=AdrTyp: DO_ADR_ADR;		\9
  NOS=ImmTyp & TOS=AdrTyp: DO_IMM_ADR;		\10

  NOS=RegTyp & TOS=ImmTyp: DO_REG_IMM;		\11
  NOS=StkTyp & TOS=ImmTyp: DO_STK_IMM;		\12
  NOS=AdrTyp & TOS=ImmTyp: DO_ADR_IMM;		\13
  NOS=ImmTyp & TOS=ImmTyp: DO_IMM_IMM		\14
other	BugMsg("GenOp2");

if NOS#ImmTyp ! TOS#ImmTyp then
	[DropV(2);  PushV(RegTyp, 0, 0)];
];	\GenOp2



proc	Common(Op);		\All Gen procedures call this
int	Op;
[LastOp:= Op;
];	\Common

\-------------------------------------------------------------------------------

proc	GenEXIT;		\$00: Exit XPL0 program
[Common($00);
Txt("	ldr	sp, [r8, #stkptr-dseg]	@00 EXIT|");
Txt("	pop	{pc}||");
];	\GenEXIT



proc	GenLOD;			\$01: TOS:= @(Level + Offset)
int	Lev, Off;
[Lev:= HexIn(3);
 Off:= HexIn(3);
Common($01);
if Lev=RegLev(0) & Off=RegOff(0) then	\value of the variable is already in r0
	PushV(RegTyp, 0, 0)
else	PushV(AdrTyp, Lev, Off);	\store variable on virtual stack
];	\GenLOD



proc	GenLDX;			\$02: TOS_byte:= @(NOS + TOS)
[Common($02);			\e.g: Byte:= A(3); A=NOS, 3=TOS
GenOp2("	ldrb	r0, [&]		@02 LDX|", false, false, "+");
RegLev(0):= -1;  RegImm(0):= false;
];	\GenLDX



proc	GenSTO;			\$03: @(Level + Offset):= TOS
\Generate code to store the value in the top of virtual stack into a variable
int	Lev, Off, I, Rs, Rd;	\source register and destination register
[Lev:= HexIn(3);
 Off:= HexIn(3);
Common($03);
\Delete register variable before calling OpString to store into the variable,
\ else it will store into the register rather than into its memory location.
\ Also, expressions like N:= N+1 would leave two versions of N in register
\ variables, one being the old value. Regs r5 and r6 are normally not bombed.
for I:= 0 to RegMax do
	if Lev=RegLev(I) & Off=RegOff(I) then RegLev(I):= -1;

if AdrTyp#VSTyp(VSP-1) ! Lev#VSLev(VSP-1) ! Off#VSOff(VSP-1) then \dont gen A:=A
   [Rs:= GetReg(0);
   TOS2Reg(Rs);
   Rd:= RegVar(Lev, Off);
   if Rd >= 0 then	\Rd is a register variable for a declared array
	Txt("	mov	r$, r%			@03 STO||", Rs, Rd)
   else	[OpString("	str	r%, &		@03 STO||", ^A, Lev, Off, Rs);
	if Rs <= RegMax then [RegLev(Rs):= Lev;  RegOff(Rs):= Off];
	];		\don't change values of registers for declared arrays
   ];
DropV(1);
];	\GenSTO



\(unused)
proc	GenSTX;			\$04: Store byte in TOS at address in NOS
[Common($04);
];	\GenSTX



proc	GenSTX2;		\$E5: @(NOS2+NOS):= TOS_byte
int	Rt, Rn, Rn2;
[Common($E5);
Rt:= GetReg(0);
TOS2Reg(Rt);
DropV(1);
Rn:= GetReg(1<<Rt);
TOS2Reg(Rn);
DropV(1);
Rn2:= GetReg(1<<Rt ! 1<<Rn);
TOS2Reg(Rn2);
DropV(1);
Txt("	strb	r%, [r?, r$]		@E5 STX2||", Rt, Rn, Rn2);
];	\GenSTX2



proc	SymOut(S);
char	S;
[];



proc	GenCAL;			\$05: Procedure call
int	Label, Ext, I;		\flag: external procedure (eproc or efunc)
int 	Lev;
[Label:= HexIn(3);
Common($05);
CleanReg(0);
for I:= 0 to RegMax do		\bomb registers possibly used by by procedure
	[RegLev(I):= -1;  RegImm(I):= false];
RegLastUsed:= 0;
if RegLev(7) >= 0 then		\save r7 if in a 'for' loop
	[Lev:= RegLev(7);  RegLev(7):= -1;
	OpString("	str	r7, &|", ^A, Lev, RegOff(7), 0);
	RegLev(7):= Lev;
	];
if FSP=9 then Txt("	vpush	{d8}|");
if FSP>9 then Txt("	vpush	{d8-d%}|", FSP-1);
Txt("	bl	K");
\\if Ext then [ChOut(OD, ^_);  SymOut(Label)]
\\else [Txt("K");
IntOut(OD, Label);
Txt("			@05 CAL|");
if FSP=9 then Txt("	vpop	{d8}|");
if FSP>9 then Txt("	vpop	{d8-d%}|", FSP-1);
OldLev:= -1;			\called routine could have altered r12
if RegLev(7) >= 0 then		\restore r7 if in a 'for' loop
	[Lev:= RegLev(7);  RegLev(7):= -1;
	OpString("	ldr	r7, &|", ^A, Lev, RegOff(7), 0);
	RegLev(7):= Lev;
	];
];	\GenCAL



proc	GenRET;			\$06: Return from a procedure
int	Lev;
[Lev:= HexIn(3);
if LastOp=$00\EXIT\ ! LastOp=$06\RET\ ! LastOp=$07\JMP\ then return;

Common($06);

\\if Lev = 0 then
\\	[Txt("	pop	{pc}			@06 RET||");  return];

if NeedBase & 1<<Lev then
	[\Restore base address of local variables at current level
	\Also restore heap pointer (r9), and r11 for calling procedure
	\Don't alter r0 because a value might be returned in it
	\A space character after "ldr " prevents re-scheduling ldr before pop.
	Txt("	pop	{r11}			@06 RET|");
	Txt("	ldr 	r9,  [r8, #base%-dseg]|", Lev);
	Txt("	str	r11, [r8, #base%-dseg]|", Lev);
	Txt("	pop	{r11, pc}||");		\restore r11 and return
	]
else	[Txt("	mov	r9, r11			@06 RET|");
	Txt("	pop	{r11, pc}||");
	];
\\	if NeedBase & 1<<Lev then ChIn(1);	\debug
\\debug	NeedBase:= NeedBase & ~1<<Lev;		\clear flag for current level
];	\GenRET



proc	GenJMP;			\$07: Jump to label
int	Label;
[Label:= HexIn(3);
if LastOp=$00\EXIT\ ! LastOp=$06\RET\ ! LastOp=$07\JMP\ !
	LastOp=$27\RTS\ then return;
Common($07);
Txt("	b	K%			@07 JMP||", Label);
];	\GenJMP



proc	GenJPC;			\$08: Jump on false (0)
int	Sense, Label;		\Sense is normally false
[Sense:= HexIn(3);
 Label:= HexIn(3);
Common($08);
CleanReg(1);
TOS2Reg(0);
Txt("	cmp	r0, #0			@08 JPC|");
Txt(if Sense then "	bne	K%|" else "	beq	K%|", Label);
DropV(1);
];	\GenJPC



proc	GenHPI;			\$09: hp:= hp + Bytes
int	Bytes;
[Bytes:= HexIn(3);
Common($09);
if Bytes>>24 then Error(1);
if Bytes & $FF then
	[Txt("	add	r9, #0x0000");  HexB(Bytes);
	Txt("		@09 HPI|")];
if Bytes>>8 & $FF then
	[Txt("	add	r9, #0x00");  HexB(Bytes>>8);  Txt("00|")];
if Bytes>>16 then Error(1);
];	\GenHPI



proc	GenARG;			\$0A: Copy proc args from virtual stack to heap
int	Bytes, Items, I;
[Bytes:= HexIn(3);
Items:= Bytes/IntSize;		\number of integers to be passed as arguments
Common($0A);
for I:= 1 to Items do
	[TOS2R0;
	Txt("	str	r0, [r9, #%]		@0A ARG|", Bytes-(IntSize*I));
	DropV(1);
	];
];	\GenARG



proc	GenIMM;			\$0B & $24 (both 32-bit and 8-bit):
int	T,	\type:	^L=label	^C=constant
	V;	\value:	label no.	immediate
[T:= HexIn(3);
 V:= HexIn(3);
Common($0B);
PushV(ImmTyp, V, T);
];	\GenIMM



proc	GenCML;			\$0C: Call intrinsic (Machine Language routine)
def	ININT=15;	\INTEGER INTRINSIC ID
def	RLINT=16;	\REAL INTRINSIC ID
int	Label,	\number of the label where the call jumps to
	Last,	\factor type (integer or real) of the last (rightmost) argument
	Type,	\intrinsic 'code' type (returns an integer or a real)
	Func,	\flag: 'true' if intrinsic is a function that returns a value
	Args,	\number of arguments (integers or reals)
	I;
def	INTEGER=1;
[Label:= HexIn(3);
 Last:= HexIn(3);
 Type:= HexIn(3);
 Func:= HexIn(3);
 Args:= HexIn(3);
Common($0C);


cond false;
if Type=INTEGER & Last#INTEGER ! Type#INTEGER & Args=2 then
	[FSP:= FSP-1;		\Fix, RlOut, ATan2, Mod
	Txt("	vpush	{d%}|", FSP);
	];
if Type#INTEGER & Last=INTEGER & Func then
	[FSP:= FSP+1;		\RlRes,  RlIn, Float
	];
cond true;


case Label of
  46,
  47:	[];						\X:= RlIn(0)
  48:	[FSP:= FSP-1;  Txt("	vpush	{d%}|", FSP)];	\RlOut(0, X)
  49:	[];						\X:= Float(N)
  50:	[FSP:= FSP-1;  Txt("	vpush	{d%}|", FSP)];	\N:= Fix(X)
  51,53,54,55,59,61,62,63,64,65,
  56,60:[FSP:= FSP-1;  Txt("	vpush	{d%}|", FSP)];	\X:= Sin(X) Cos(X)
  66,
  57,58:[FSP:= FSP-2;  Txt("	vpush	{d%}|", FSP);	\X:= ATan2(Y,X) Mod  Y
	 	       Txt("	vpush	{d%}|", FSP+1);	\		     X
	]
other	[];


if Args > 0 then
	[if Type=ININT then	\handle all combinations of arguments
		[if Last=INTEGER then TOS2R0 else Txt("	pop	{r0}|")]
	else	TOS2Stack;
	];
Txt("	bl	intr%			@0C CML|", Label);

DropV(Args);
OldLev:= -1;			\intrinsic call could have altered r12
for I:= 0 to 3 do		\bomb registers used by intrinsic
	[RegLev(I):= -1;  RegImm(I):= false];
RegLastUsed:= 0;

if Func then
	case Type of
	  ININT: PushV(RegTyp, 0, 0);
	  RLINT: PushV(RealTyp, 0, 0)
	other	 [];

case Label of
  46,
  47:	[Txt("	vpop	{d%}|", FSP);  FSP:= FSP+1];	\X:= RlIn(0)
  48:	[];						\RlOut(0, X)
  49:	[Txt("	vpop	{d%}|", FSP);  FSP:= FSP+1];	\X:= Float(N)
  50:	[];						\N:= Fix(X)
  51,53,54,55,59,61,62,63,64,65,
  56,60:[Txt("	vpop	{d%}|", FSP);  FSP:= FSP+1];	\X:= Sin(X) Cos(X)
  66,
  57,58:[Txt("	vpop	{d%}|", FSP);  FSP:= FSP+1]	\X:= ATan2(Y,X) Mod
other	[];
];	\GenCML



proc	GenADD;			\$0D: TOS:= NOS+TOS
int	Cons;	\both operands are constants, yielding a constant result
[Common($0D);
if not Identity(0, true) then
	[Cons:= VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C &
		VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C;
	GenOp2("	add	r0, &		@0D ADD|", false, false, "+");
	if not Cons then
		[RegLev(0):= -1;  RegImm(0):= false];
	];
];	\GenADD



proc	GenSUB;			\$0E: TOS:= NOS-TOS
int	Cons;	\both operands are constants, yielding a constant result
[Common($0E);
if not Identity(0, false) then
	[Cons:= VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C &
		VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C;
	GenOp2("	sub	r0, &		@0E SUB|", true, false, "-");
	if not Cons then
		[RegLev(0):= -1;  RegImm(0):= false];
	];
];	\GenSUB



func	Power2(N);		\Return N's power of 2 if it's exact & positive
int	N;			\ e.g: if N=8 then return 3;  if N=9 return 0
int	P;			\Power of 2
[if N <= 0 then return 0;
P:= 30;
while N do
	[N:= N<<1;
	if N < 0 then
		return if N<<1 then 0 else P;
	P:= P-1;
	];
];	\Power2



proc	GenMUL;			\$0F: TOS:= NOS*TOS
int	TOSTyp, NOSTyp, N,
	Cons;	\both operands are constants, yielding a constant result
[Common($0F);
if not Identity(1, true) then
    if not IsZero(true) then
	[TOSTyp:= VSTyp(VSP-1);
	 NOSTyp:= VSTyp(VSP-2);
	 N:= Power2(VSLev(VSP-1));
	 if TOSTyp=ImmTyp & N#0 & VSOff(VSP-1)=^C &
	    NOSTyp#ImmTyp then	\optimize A*PowerOf2
		[DropV(1);
		CleanReg(1);
		TOS2Reg(0);
		Txt("	lsl	r0, #%|", N);
		RegLev(0):= -1;  RegVal(0):= RegVal(0)<<N;
		DropV(1);
		PushV(RegTyp, 0, 0);
		]
	else if NOSTyp=ImmTyp & Power2(VSLev(VSP-2))#0 & VSOff(VSP-2)=^C &
	    TOSTyp#ImmTyp then
		[N:= Power2(VSLev(VSP-2));	\optimize PowerOf2*A
		VSTyp(VSP-2):= VSTyp(VSP-1);
		VSLev(VSP-2):= VSLev(VSP-1);
		VSOff(VSP-2):= VSOff(VSP-1);
		DropV(1);
		CleanReg(1);
		TOS2Reg(0);
		Txt("	lsl	r0, #%|", N);
		RegLev(0):= -1;  RegVal(0):= RegVal(0)<<N;
		DropV(1);
		PushV(RegTyp, 0, 0);
		]
	else	[Cons:= VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C &
			VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C;
		GenOp2("	mul	r0, &		@0F MUL|",
			false, true, "*");
		if not Cons then
			[RegLev(0):= -1;  RegImm(0):= false];
		];
	];
];	\GenMUL



proc	GenDIV;			\$10: TOS:= NOS/TOS; Remainder
int	I;
[Common($10);
\Don't call Identity and IsZero both at same time because they can alter VStack.
if Identity(1, false) then			\A/1=A; rem=0
	[CleanReg(0);
	if ~RegImm(0) ! RegVal(0)#0 then Txt("	mov	r0, #0|");
	Txt("	str	r0, [r8, #remain-dseg]|");
	RegLev(0):= -1;  RegVal(0):= 0;  RegImm(0):= true;
	]
else if IsZero(false) then			\0/A=0; rem=0
	[CleanReg(0);
	if ~RegImm(0) ! RegVal(0)#0 then Txt("	mov	r0, #0|");
	Txt("	str	r0, [r8, #remain-dseg]|");
	RegLev(0):= -1;  RegVal(0):= 0;  RegImm(0):= true;
	]
else if VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C &
	VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C then
		[Txt("					@10 DIV const|");
		GenOp2("@10 DIV|", true, true, "/")
		]
else	[TOS2R0;
	DropV(1);
	if VSTyp(VSP-1)#StkTyp then		\if not already on the stack
		[TOS2Reg(1);  Txt("	push	{r1}|")];
	Txt("	bl	sdivx			@10 DIV|");	\r0:= tos/r0
	for I:= 0 to 3 do			\bomb sdivx subroutine registers
		[RegLev(I):= -1;  RegImm(I):= false];
	DropV(1);
	PushV(RegTyp, 0, 0);
	];
];	\GenDIV



proc	GenNEG;			\$11: TOS:= -TOS
[Common($11);
if VSTyp(VSP-1)=ImmTyp then
	VSLev(VSP-1):= -VSLev(VSP-1)
else	[CleanReg(1);
	TOS2Reg(0);
	Txt("	rsb	r0, #0			@11 NEG|");
	RegLev(0):= -1;  RegVal(0):= -RegVal(0);
	DropV(1);
	PushV(RegTyp, 0, 0);
	];
];	\GenNEG



proc	GenCMP(Op);		\$12-$17: Integer compares:  TOS:= NOS>TOS
int	Op;
int	TOSTyp, NOSTyp, CmpTbl;
[Common(Op);
TOSTyp:= VSTyp(VSP-1);
NOSTyp:= VSTyp(VSP-2);
CleanReg(2);
CmpTbl:= ["=", "#", ">=", ">", "<=", "<"];
GenOp2("	subs	r0, &|", true, false, CmpTbl(Op-$12));
\"cmp r0, r3, r0" is invalid syntax, so "subs" is used instead
if TOSTyp#ImmTyp ! NOSTyp#ImmTyp then
	[CmpTbl:= [
	"	subeq	r0, #1			@12 CMP|",
	"	subne	r0, #1			@13 CMP|",
	"	subge	r0, #1			@14 CMP|",
	"	subgt	r0, #1			@15 CMP|",
	"	suble	r0, #1			@16 CMP|",
	"	sublt	r0, #1			@17 CMP|"];
	Txt("	mov	r0, #0|");	\assume 'false' (0)
	Txt(CmpTbl(Op-$12));		\change to 'true' (-1) by subtracting 1
	];
RegLev(0):= -1;  RegImm(0):= false;
];	\GenCMP



proc	GenJOC(Op);		\$12-$17: Jump if NOS > TOS, etc.
int	Op;
int	Sense, Label, CmpTbl, TOSTyp, NOSTyp;
[Sense:= HexIn(3);		\unused
 Label:= HexIn(3);
Common(Op!$100);		\distinguish from CMP
CleanReg(2);
TOSTyp:= VSTyp(VSP-1);
NOSTyp:= VSTyp(VSP-2);
Txt("					@12..17 JOC|");
if Sense then
	CmpTbl:= ["#", "=", "<", "<=", ">", ">="]
else	CmpTbl:= ["=", "#", ">=", ">", "<=", "<"];
GenOp2("	subs	r14, &|", true, false, CmpTbl(Op-$12));
\"cmp r14, r2, r3" is invalid syntax, so "subs" is used instead
if TOSTyp=ImmTyp & NOSTyp=ImmTyp then
	[if VSLev(VSP-1) = Sense then
		[Txt("	b	K%|", Label);
		Common($07\JMP\);	\possibly eliminates dead code
		];
	]
else	[if Sense then
		CmpTbl:= [
		"	beq	K%|",
		"	bne	K%|",
		"	bge	K%|",
		"	bgt	K%|",
		"	ble	K%|",
		"	blt	K%|"]
	else	CmpTbl:= [
		"	bne	K%|",
		"	beq	K%|",
		"	blt	K%|",
		"	ble	K%|",
		"	bgt	K%|",
		"	bge	K%|"];
	Txt(CmpTbl(Op-$12), Label);
	];
DropV(1);
];	\GenJOC



proc	GenFOR(DownTo);		\$18, $49: if NOS>=TOS then loop
int	DownTo;			\	   if [sp]>=r7 then loop
int	Label, Lev, Off, I;
[Label:= HexIn(3);
Lev:= HexIn(3);
Off:= HexIn(3);
Common(if DownTo then $49 else $18);
Txt(if DownTo then "					@49 FORDN|"
	       else "					@18 FOR|");
Txt("	ldr	r0, [sp]|");	\get limit value

\Make sure that the control variable in memory is correct when exiting the loop
RegLev(7):= -1;						\avoid "str r7, r7"
for I:= 0 to RegMax do					\avoid "str r7, r3" etc.
	if Lev=RegLev(I) & Off=RegOff(I) then RegLev(I):= -1;
OpString("	str	r7, &|", ^A, Lev, Off, 0);

Txt("	cmp	r7, r0|");
Txt(if DownTo then "	bge	K%|" else "	ble	K%|", Label);
Txt("	add	sp, #4|");	\discard limit value
DropV(1);

if ForSP > 0 then		\restore r7 if coming out of a nested 'for' loop
	[Lev:= ForStkLev(ForSP-1);
	 Off:= ForStkOff(ForSP-1);
	 OpString("	ldr	r7, &|", ^A, Lev, Off, 0);
	 RegLev(7):= Lev;  RegOff(7):= Off;
	 DropFor;
	];
CrLf(OD);
];	\GenFOR



proc	GenINP(DownTo);		\$19, $4A: Increment control variable & push it
int	DownTo;
int	Lev, Off;
[Lev:= HexIn(3);
 Off:= HexIn(3);
Common(if DownTo then $4A else $19);
if Lev#RegLev(7) ! Off#RegOff(7) then	\if ctrl var is not in r7 then load it
	OpString("	?	r7, &|", ^A, Lev, Off, 0);
if DownTo then
	Txt("	sub	r7, #1			@4A DEP|")
else Txt("	add	r7, #1			@19 INP|");
];	\GenINP



proc	GenOR;			\$1A: TOS:= NOS!TOS
int	Cons;	\both operands are constants, yielding a constant result
[Common($1A);
if not Identity(0, true) then
	[Cons:= VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C &
		VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C;
	GenOp2("	orr	r0, &		@1A OR|", false, false, "!");
	if not Cons then
		[RegLev(0):= -1;  RegImm(0):= false];
	];
];	\GenOR



proc	GenAND;			\$1B: TOS:= NOS&TOS
int	Cons;	\both operands are constants, yielding a constant result
[Common($1B);
if not Identity($FFFF_FFFF, true) then
    if not IsZero(true) then
	[Cons:= VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C &
		VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C;
	GenOp2("	and	r0, &		@1B AND|", false, false, "&");
	if not Cons then
		[RegLev(0):= -1;  RegImm(0):= false];
	];
];	\GenAND



proc	GenNOT;			\$1C: TOS:= ~TOS
[Common($1C);
if VSTyp(VSP-1)=ImmTyp then
	VSLev(VSP-1):= ~ VSLev(VSP-1)
else	[CleanReg(1);
	TOS2Reg(0);
	Txt("	mvn	r0, r0			@1C NOT|");
	RegLev(0):= -1;  RegVal(0):= ~RegVal(0);
	DropV(1);
	PushV(RegTyp, 0, 0);
	];
];	\GenNOT



proc	GenEOR;			\$1D: TOS:= NOS|TOS
int	Cons;	\both operands are constants, yielding a constant result
[Common($1D);
if not Identity(0, true) then
	[Cons:= VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C &
		VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C;
	GenOp2("	eor	r0, &		@1D EOR|", false, false, "|");
	if not Cons then
		[RegLev(0):= -1;  RegImm(0):= false];
	];
];	\GenEOR



proc	GenDBA;			\$1E: TOS:= NOS + TOS*IntSize
int	Imm;
[Common($1E);
Imm:= VSLev(VSP-1);
if VSTyp(VSP-1)=ImmTyp & ImmOk(Imm*IntSize) & VSOff(VSP-1)=^C then
	[DropV(1);		\A(3):=
	TOS2R0;  DropV(1);	\nos
	if Imm \#0\ then
		Txt("	add	r0, #%  		@1E DBA|", Imm*IntSize);
	PushV(RegTyp, 0, 0);
	]
else	GenOp2("	add	r0, &, lsl #2	@1E DBA|", true, true, "+");
RegLev(0):= -1;  RegImm(0):= false;
];	\GenDBA



proc	GenSTD;			\$1F: @(NOS):= TOS
int	Rt, Rn;
[Common($1F);
Rt:= GetReg(0);
TOS2Reg(Rt);
DropV(1);
Rn:= GetReg(1<<Rt);
TOS2Reg(Rn);
DropV(1);
Txt("	str	r%, [r$]		@1F STD||", Rt, Rn);
];	\GenSTD



proc	GenDBX;			\$20: TOS:= @(NOS + TOS<<2)
int	Imm;
[Common($20);			\e.g: B:= A(3); A=NOS, 3=TOS
Imm:= VSLev(VSP-1);
if VSTyp(VSP-1)=ImmTyp & ImmOk(Imm*IntSize) & VSOff(VSP-1)=^C then
	[DropV(1);		\A(3)
	TOS2R0;  DropV(1);	\nos
	Txt("	ldr	r0, [r0, #%] 		@20 DBX|", Imm*IntSize);
	PushV(RegTyp, 0, 0);
	]
else	GenOp2("	ldr	r0, [&, lsl #2]	@20 DBX|", true, true, "+");
RegLev(0):= -1;  RegImm(0):= false;
];	\GenDBX



proc	GenADROff(Off, Reg);	\Generate offset for ADR op
int	 Off, Reg;
[Txt("	add	r0, r%, #$|", Reg, Off&$0000_00FF);
if Off & $0000_FF00 then
	Txt("	add	r0, #%|", Off&$0000_FF00);
if Off & $00FF_0000 then
	Txt("	add	r0, #%|", Off&$00FF_0000);
if Off & $FF00_0000 then
	Txt("	add	r0, #%|", Off&$FF00_0000);
RegLev(0):= -1;  RegImm(0):= false;
];	\GenADROff



proc	GenADR;			\$21: Load the address of a variable
int	Lev, Off, I;
[Lev:= HexIn(3);
 Off:= HexIn(3);
Common($21);
Txt("					@21 ADR|");
\Beware of a pointer to a variable. It could be used to store into the variable
\ while its shadowing register remains unchanged. This prevents some instances.
for I:= 0 to RegMax do
	if Lev=RegLev(I) & Off=RegOff(I) then RegLev(I):= -1;
if Lev=RegLev(7) & Off=RegOff(7) then RegLev(7):= -1;

CleanReg(0);
case Lev of
  0:	GenADROff(Off, 10);
  Level:GenADROff(Off, 11)
other	[NewLevel(Lev);  GenADROff(Off, 12)];
PushV(RegTyp, 0, 0)
];	\GenADR



proc	GenFADR;		\$22: FTOS:= Base(Level)+Offset
int	Lev, Off;		\Load the address of a real variable
[Lev:= HexIn(3);
 Off:= HexIn(3);
Common($22);
Txt("					@22 FADR|");

CleanReg(0);
case Lev of
  0:	GenADROff(Off, 10);
  Level:GenADROff(Off, 11)
other	[NewLevel(Lev);  GenADROff(Off, 12)];

Txt("	mov	r14, #0x3F00|");
Txt("	orr	r14, #0x00F0|");
Txt("	fmdrr	d%, r0, r14|", FSP);
FSP:= FSP+1;
RegLev(0):= -1;  RegImm(0):= false;
PushV(RealTyp, 0, 0);
];	\GenFADR



proc	GenLDA;			\$23: (unused)
[Common($23);
Txt("					@23 LDA|");
Error(2);
];	\GenLDA

				\$24: (see $0B)

proc	GenCJP;			\$25: Conditional jump, i.e: if TOS = NOS.
\NOS is normally in r0, and it equals the value of the expression between the
\ words 'case' & 'of'. TOS equals the "label" value, the value before the ":".
\Compare NOS to TOS and jump if they're equal. Discard TOS in either case.
int	Sense, Label;
[Sense:= HexIn(3);
 Label:= HexIn(3);
Common($25);
if VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C & ImmOk(VSLev(VSP-1)) &
   VSTyp(VSP-2)=RegTyp then					\subs r14, r0,#4
	[OpString("	subs	r14, &", ^I, VSLev(VSP-1), VSOff(VSP-1), 0);
	\"cmp r14, r0, #4" is invalid syntax, so "subs" is used instead
	Txt("		@25 CJP|");
	DropV(1);
	]
else	[TOS2Reg(1);
	DropV(1);
	TOS2R0;
	Txt("	cmp	r0, r1			@25 CJP|");
	];
Txt(if Sense then "	beq	K%|" else "	bne	K%|", Label);
];	\GenCJP



proc	GenJSR;			\$26: Optimized procedure call
int	Label, I, Lev;
[Label:= HexIn(3);
Common($26);
CleanReg(0);
for I:= 0 to RegMax do		\bomb registers possibly altered by procedure
	[RegLev(I):= -1;  RegImm(I):= false];
RegLastUsed:= 0;
if RegLev(7) >= 0 then		\if in a for loop then save r7
	[Lev:= RegLev(7);  RegLev(7):= -1;
	OpString("	str	r7, &|", ^A, Lev, RegOff(7), 0);
	RegLev(7):= Lev;
	];
if FSP=9 then Txt("	vpush	{d8}|");
if FSP>9 then Txt("	vpush	{d8-d%}|", FSP-1);
Txt("	bl	K%			@26 JSR|", Label);
if FSP=9 then Txt("	vpop	{d8}|");
if FSP>9 then Txt("	vpop	{d8-d%}|", FSP-1);
OldLev:= -1;			\procedure call could have altered r12
if RegLev(7) >= 0 then		\restore r7 if in a 'for' loop
	[Lev:= RegLev(7);  RegLev(7):= -1;
	OpString("	ldr	r7, &|", ^A, Lev, RegOff(7), 0);
	RegLev(7):= Lev;
	];
];	\GenJSR



proc	GenRTS;			\$27: Optimized procedure return
[if LastOp=$00\EXIT\ ! LastOp=$07\JMP\ ! LastOp=$27\RTS\ then return;
Common($27);
Txt("	pop	{r11, pc}		@27 RTS||");
];	\GenRTS



proc	GenDRP(N);		\$28: Drop N words from hardware stack
int	N;
int	I;
[N:= HexIn(3);
Common($28);
Txt("					@28 DRP|");
if VSP-N < 0 then BugMsg("GenDRP VSP-N < 0");
for I:= 1 to N do
	[if VSTyp(VSP-I) = StkTyp then
		Txt("	add	sp, #4|");
	];
];	\GenDRP



proc	GenCEXT(Label, Last, Type, Func, Args);	\$29: External assembly routine
int	Label, Last, Type, Func, Args, I;
def	INTEGER=1;
[Common($29);
CleanReg(0);
cond false;
if Args>0 & Last=INTEGER then TOS2Stack;
Txt("	bl	");  SymOut(Label);
Txt("			@29 CEXT|");
DropV(Args);
OldLev:= -1;			\subroutine call could have altered r12
for I:= 0 to 3 do		\bomb registers altered by subroutine
	[RegLev(I):= -1;  RegImm(I):= false];
if Func then
	case Type of
	  INEXT: PushV(StkTyp, 0, 0);
	  RLEXT: PushV(RealTyp, 0, 0)
	other;
cond true;
];	\GenCEXT



proc	GenFLOD;		\$2A: Load floating point variable
int	Lev, Off;		\ FTOS:= @(Lev+Off)

	proc	GenOff(Rn);
	int	 Rn;
	[if Off >= 4096 then Error(1);
	Txt("	fldd	d%, [r$, #?]		@2A FLOD|", FSP, Rn, Off);
	FSP:= FSP+1;
	];	\GenOff

[Lev:= HexIn(3);
 Off:= HexIn(3);
Common($2A);
case Lev of
  0:	GenOff(10);
  Level:GenOff(11)
other	[NewLevel(Lev);  GenOff(12)];
PushV(RealTyp, 0, 0);
];	\GenFLOD



proc	GenFSTO;		\$2B: Store floating point variable
int	Lev, Off;		\ @(Lev+Off):= FTOS

	proc	GenOff(Rn);
	int	 Rn;
	[if Off >= 4096 then Error(1);
	FSP:= FSP-1;
	Txt("	fstd	d%, [r$, #?]		@2B FSTO|", FSP, Rn, Off);
	];	\GenOff

[Lev:= HexIn(3);
 Off:= HexIn(3);
Common($2B);
case Lev of
  0:	GenOff(10);
  Level:GenOff(11)
other	[NewLevel(Lev);  GenOff(12)];
DropV(1);
];	\GenFSTO



proc	GenFIMM;		\$2C: FTOS:= constant
int	Typ, Val;
[Typ:= HexIn(3);
 Val:= HexIn(3);
Common($2C);
if Typ = ^C then			\Type = Constant
	[Txt("	ldr	r14, KP%  		@2C FIMM|", Val);
	Txt("	b	KK%|", Val);
	Txt("KP%:	.word	L%|", Val);
	Txt("KK%:|", Val);
	Txt("	fldd	d%, [r14]|", FSP);
	FSP:= FSP+1;
	]
else if Typ = ^L then			\Type = Label (for constant array)
	[Txt("	ldr	r0, KP%  		@2C FIMM|", Val);
	Txt("	b	KK%|", Val);
	Txt("KP%:	.word	L%|", Val);
	Txt("KK%:|", Val);
	Txt("	mov	r14, #0x3F000000|");
	Txt("	orr	r14, #0x00F00000|");
	Txt("	fmdrr	d%, r0, r14|", FSP);
	FSP:= FSP+1;
	RegLev(0):= -1;  RegImm(0):= false;
	]
else	Error(3);
PushV(RealTyp, 0, 0);
];	\GenFIMM



proc	GenFADD;		\$2D: FTOS:= FNOS+FTOS
[Common($2D);
FSP:= FSP-1;
Txt("	faddd	d%, d$, d?		@2D FADD|", FSP-1, FSP-1, FSP);
DropV(1);
];	\GenFADD



proc	GenFSUB;		\$2E: FTOS:= FNOS-FTOS
[Common($2E);
FSP:= FSP-1;
Txt("	fsubd	d%, d$, d?		@2E FSUB|", FSP-1, FSP-1, FSP);
DropV(1);
];	\GenFSUB



proc	GenFMUL;		\$2F: FTOS:= FNOS*FTOS
[Common($2F);
FSP:= FSP-1;
Txt("	fmuld	d%, d$, d?		@2F FMUL|", FSP-1, FSP-1, FSP);
DropV(1);
];	\GenFMUL



proc	GenFDIV;		\$30: FTOS:= FNOS/FTOS
[Common($30);
FSP:= FSP-1;
Txt("	fdivd	d%, d$, d?		@30 FDIV|", FSP-1, FSP-1, FSP);
DropV(1);
];	\GenFDIV



proc	GenFNEG;		\$31: FTOS:= -FTOS
[Common($31);
Txt("	fnegd	d%, d%  		@31 FNEG|", FSP-1);
];	\GenFNEG



proc	GenFCMP(Op);		\$32-$37: real compares, e.g: TOS:= FNOS>FTOS
\Compare top two items on FPU stack and return true (-1) or false (0) in r0
int	Op;
int	CMPT;
[Common(Op);
CMPT:= [
"	subeq	r0, #1			@32 FCMP|",
"	subne	r0, #1			@33 FCMP|",
"	subge	r0, #1			@34 FCMP|",
"	subgt	r0, #1			@35 FCMP|",
"	suble	r0, #1			@36 FCMP|",
"	sublt	r0, #1			@37 FCMP|"];
CleanReg(0);
FSP:= FSP-2;
Txt("	fcmpd	d%, d$|", FSP, FSP+1);
Txt("	mov	r0, #0|");
Txt("	fmstat|");
Txt(CMPT(Op-$32));
RegLev(0):= -1;  RegImm(0):= false;
DropV(2);
PushV(RegTyp, 0, 0);
];	\GenFCMP



proc	GenTRA;			\$38: TOS:= NOS + TOS*8
int	Imm;
[Common($38);
Imm:= VSLev(VSP-1);
if VSTyp(VSP-1)=ImmTyp & ImmOk(Imm*8) & VSOff(VSP-1)=^C then
	[DropV(1);		\A(3):=
	TOS2R0;  DropV(1);	\nos
	if Imm \#0\ then
		Txt("	add	r0, #%  		@38 TRA|", Imm*8);
	PushV(RegTyp, 0, 0);
	]
else	GenOp2("	add	r0, &, lsl #3	@38 TRA|", true, true, "+");
RegLev(0):= -1;  RegImm(0):= false;
];	\GenTRA



proc	GenTRX;			\$39: TOS:= @(NOS + TOS*8)
int	Imm;
[Common($39);
Imm:= VSLev(VSP-1);
if VSTyp(VSP-1)=ImmTyp & ImmOk(Imm*8) & VSOff(VSP-1)=^C then
	[DropV(1);		\A(3,N)
	TOS2R0;  DropV(1);	\nos
	Txt("	ldr	r0, [r0, #%] 		@39 TRX|", Imm*8);
	PushV(RegTyp, 0, 0);
	]
else	GenOp2("	ldr	r0, [&, lsl #3]	@39 TRX|", true, true, "+");
RegLev(0):= -1;  RegImm(0):= false;
];	\GenTRX



proc	GenTRI;			\$3A: FTOS:= @(FNOS + TOS*8)
int	Imm;			\e.g: A(3);  A=FNOS, 3=TOS
[Common($3A);
\Fetch address from low half of (real) FNOS
Txt("	fmrdl	r14, d% 		@3A TRI|", FSP-1);
Imm:= VSLev(VSP-1);
if VSTyp(VSP-1)=ImmTyp & ImmOk(Imm*8) & VSOff(VSP-1)=^C then
	[DropV(1);		\A(3)
	if Imm \#0\ then
		Txt("	add	r14, #%|", Imm*8);
	]
else if VSTyp(VSP-1)=AdrTyp & VStkIsReg(1, 7) then
	[DropV(1);		\A(r7)
	Txt("	add	r14, r7, lsl #3|");
	]
else	[TOS2R0;  DropV(1);	\A(N)
	Txt("	lsl	r0, #3|");
	RegLev(0):= -1;  RegVal(0):= RegVal(0)<<3;
	Txt("	add	r14, r0|");
	];
Txt("	fldd	d%, [r14]|", FSP-1);	\replace FTOS
];	\GenTRI



proc	GenSTT;			\$3B: @(NOS):= FTOS
\Store the 8-byte real in TOS into the absolute address pointed to by NOS.
[Common($3B);
DropV(1);
TOS2Reg(14);
DropV(1);
FSP:= FSP-1;
Txt("	fstd	d%, [r14]		@3B STT|", FSP);
];	\GenSTT



proc	GenMARK;		\$3C: (unused)
[Common($3C);
Txt("					@3C MARK|");
];	\GenMARK



proc	GenASR;			\$3D: TOS:= NOS->>TOS
int	Cons;	\both operands are constants, yielding a constant result
[Common($3D);
if not Identity(0, false) then		\N->>0 = N
    if not IsZero(false) then		\0->>N = 0
	[Cons:= VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C &
		VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C;
	GenOp2("	asr	r0, &		@3D ASR|", true, false, "->>");
	if not Cons then
		[RegLev(0):= -1;  RegImm(0):= false];
	];
];	\GenASR



proc	GenLSL;			\$3E: TOS:= NOS<<TOS
int	Cons;	\both operands are constants, yielding a constant result
[Common($3E);
if not Identity(0, false) then		\N<<0 = N
    if not IsZero(false) then		\0<<N = 0
	[Cons:= VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C &
		VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C;
	GenOp2("	lsl	r0, &		@3E LSL|", true, false, "<<");
	if not Cons then
		[RegLev(0):= -1;  RegImm(0):= false];
	];
];	\GenLSL



proc	GenLSR;			\$3F: TOS:= NOS>>TOS
int	Cons;	\both operands are constants, yielding a constant result
[Common($3F);
if not Identity(0, false) then		\N>>0 = N
    if not IsZero(false) then		\0>>N = 0
	[Cons:= VSTyp(VSP-1)=ImmTyp & VSOff(VSP-1)=^C &
		VSTyp(VSP-2)=ImmTyp & VSOff(VSP-2)=^C;
	GenOp2("	lsr	r0, &		@3F LSR|", true, false, ">>");
	if not Cons then
		[RegLev(0):= -1;  RegImm(0):= false];
	];
];	\GenLSR

\ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc	GenARY;			\$48: Array declaration
int	Lev, I;
[Lev:= HexIn(3);
Common($48);
CleanReg(0);			\(for safety)
Txt("	bl	mkarray			@48 ARY|");
\mkarray pulls arguments from the h/w stack and sets up a pointer in the array's
\ variable name to point to the array space. It also increases the heap pointer
\ (r9) to skip over the array space, thus reserving it.

if RegVar(Lev, VSLev(VSP-1)) >= 0 then	\set up register for fast access
	[OpString("	?	&, ", ^A, Lev, VSLev(VSP-1), 0);
	Txt("r0|");		\e.g: mov r6, r0
	];
I:= VSLev(VSP-3);		\number of dimensions (usually 1)
DropV(3+I);			\drop at least 4 items + any extra dimensions
for I:= 0 to 3 do		\bomb registers altered by mkarray subroutine
	[RegLev(I):= -1;  RegImm(I):= false];
\(Don't bomb r12, which is used by BASEARY to save the return address)
];	\GenARY

				\$49: (see $18: FOR down)
				\$4A: (see $19: INP down)

proc	GenREM;			\$4B: TOS:= Remainder of last integer divide
[Common($4B);
DropV(1);			\discard TOS
CleanReg(0);
Txt("	ldr	r0, [r8, #remain-dseg]	@4B REM|");
RegLev(0):= -1;  RegImm(0):= false;
PushV(RegTyp, 0, 0);
];	\GenREM



proc	GenSWAP;		\$4C: Swap-bytes function
[Common($4C);
if VSTyp(VSP-1)=ImmTyp then
	VSLev(VSP-1):= swap(VSLev(VSP-1))
else	[CleanReg(1);
	TOS2Reg(0);
	Txt("	rev16	r0, r0			@4C SWAP|");
	RegLev(0):= -1;  RegVal(0):= swap(RegVal(0));
	DropV(1);
	PushV(RegTyp, 0, 0);
	];
];	\GenSWAP



proc	GenEXT;			\$4D: Sign-extend function
[Common($4D);
if VSTyp(VSP-1)=ImmTyp then
	VSLev(VSP-1):= extend(VSLev(VSP-1))
else	[CleanReg(1);
	TOS2Reg(0);
	Txt("	sxtb	r0, r0			@4D EXT|");
	RegLev(0):= -1;  RegVal(0):= extend(RegVal(0));
	DropV(1);
	PushV(RegTyp, 0, 0);
	];
];	\GenEXT



proc	GenPIN;			\$4E: Function to read byte from a port
[Common($4E);			\TOS:= port(TOS)
Txt("					@4E PIN|");
Error(2);
];	\GenPIN



proc	GenPOUT;		\$4F: Function to write a byte to a port
[Common($4F);
Txt("					@4F POUT|");
Error(2);
];	\GenPOUT



proc	GenFIX;			\$50: Fix function
[Common($50);			\integer:= fix(real)
CleanReg(0);
FSP:= FSP-1;
Txt("	ftosid	s2, d%  		@50 FIX|", FSP);
Txt("	fmrs	r0, s2|");
RegLev(0):= -1;  RegImm(0):= false;
DropV(1);			\drop real from virtual stack
PushV(RegTyp, 0, 0);		\make virtual stack reflect value in r0
];	\GenFIX



proc	GenFLOAT;		\$51: Float function
[Common($51);			\real:= float(integer)
TOS2R0;
Txt("	fmsr	s0, r0			@51 FLOAT|");
Txt("	fsitod	d%, s0|", FSP);
FSP:= FSP+1;
DropV(1);		\drop integer from virtual stack
PushV(RealTyp, 0, 0);	\make virtual stack reflect value on FPU stack
];	\GenFLOAT



proc	GenFSQRT;		\$52: Square root function of a real
[Common($52);
Txt("	fsqrtd	d%, d%|			@52 FSQRT|", FSP-1);
];	\GenFSQRT



proc	GenSQRT;		\$53: Square root of an integer
[Common($53);
if VSTyp(VSP-1) = ImmTyp then
	VSLev(VSP-1):= sqrt(VSLev(VSP-1))
else	[CleanReg(1);
	TOS2Reg(0);
	Txt("	fmsr	s2, r0			@53 SQRT|");
	Txt("	fsitod	d0, s2|");		\float
	Txt("	fsqrtd	d0, d0|");
	Txt("	ftosid	s2, d0|");		\fix
	Txt("	fmrs	r0, s2|");
	RegLev(0):= -1;  RegVal(0):= sqrt(RegVal(0));
	DropV(1);
	PushV(RegTyp, 0, 0);
	];
];	\GenSQRT



proc	GenFSQ;			\$54: Square a real
[Common($54);
Txt("	fmuld	d%, d%, d%		@54 FSQ|", FSP-1);
];	\GenFSQ



proc	GenSQ;			\$55: Square an integer
[Common($55);
if VSTyp(VSP-1) = ImmTyp then
	VSLev(VSP-1):= sq(VSLev(VSP-1))
else	[CleanReg(1);
	TOS2Reg(0);
	Txt("	mul	r0, r0, r0		@55 SQ|");
	RegLev(0):= -1;  RegVal(0):= RegVal(0)*RegVal(0);
	DropV(1);
	PushV(RegTyp, 0, 0);
	];
];	\GenSQ

\ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc	GenXR7;			\$E9: Bomb R7 tracking
[Common($E9);			\ (used after 'loop' inside a 'for' statement)
Txt("					@E9 XR7|");
RegLev(7):= -1;
];	\GenXR7



proc	GenVPSHADR;		\$F0: Push AdrTyp onto virtual stack
[Common($F0);			\ (used for 'case' statements)
Txt("					@F0 VPSHADR|");
PushV(AdrTyp, 0, 0);
];	\GenVPSHADR



func	GetAvailReg;		\Return the number of an unused global register
int	I;
[for I:= 6 downto 5 do
	if RegLev(I)<0 & RegImm(I)=false then return I;
return -1;
];	\GetAvailReg



proc	GenREGARY;		\$F6: Assign register to array name of Lev & Off
int	Lev, Off, R;
\This is done near the array declaration so any following (nested) procedures
\ can use the register instead of having to load the array name from memory.
[Lev:= HexIn(3);
 Off:= HexIn(3);
Common($F6);
R:= GetAvailReg;

\Only do this for level = 0, because declared arrays in nested procedures don't
\ relinquish the register when they terminate.
if Lev=0 & R>=0 then
	[Txt("					@F6 REGARY|");
	RegLev(R):= Lev;
	RegOff(R):= Off;
	RegMax:= R-1;
	];
];	\GenREGARY



proc	GenCLEAN;		\$ED: Make sure r0 is on the h/w stack
[Common($ED);			\ (required for KeyHit intrinsic, etc.)
Txt("					@ED CLEAN|");
CleanReg(0);
];	\GenCLEAN



proc	GenBASEARY;		\$F5: Set baseary to base address of any arrays
[Common($F5);			\mkarray needs this to access array names, so it
				\ can set pointers to their array spaces
Txt("	str	r9, [r8, #baseary-dseg]	@F5 BASEARY|");
Txt("	mov	r12, r14|");	\save proc's lr cuz call to mkarray changes it
];	\GenBASEARY		(this must be done even if no calls to mkarray)



proc	GenVPSH;		\$EF: Push RegTyp onto virtual stack
[Common($EF);			\ (used for 'case' statements)
Txt("					@EF VPSH|");
PushV(RegTyp, 0, 0);
];	\GenVPSH



proc	GenREAL;		\$F2: Convert top two items on stack to a real
[Common($F2);			\converts 0, adr to a real; required for @Real
Txt("					@F2 REAL|");
DropV(2);
PushV(RealTyp, 0, 0);
];	\GenREAL



proc	GenVDRP;		\$F1: Drop one item from virtual stack
[Common($F1);			\required for 'if' expressions
Txt("					@F1 VDRP|");
DropV(1);
if FSP>8 then FSP:= FSP-1;	\kludge for 'real' 'if' expressions
];	\GenVDRP



proc	GenREG;			\$EB: Put integer TOS into r0
[Common($EB);			\required for 'if' expressions
Txt("					@EB REG|");
if VSTyp(VSP-1) # RealTyp then
	TOS2R0;
];	\GenREG



proc	GenSTK;			\$EC: Push intrinsic argument onto h/w stack
[Common($EC);
Txt("					@EC STK|");
TOS2Stack;
];	\GenSTK



proc	GenFORSET;		\$E8: initialize a 'for' loop
int	Lev;
[Common($E8);
Txt("					@E8 FORSET|");

if RegLev(7) >= 0 then		\save r7 if in a 'for' loop
	[Lev:= RegLev(7);  RegLev(7):= -1;
	OpString("	str	r7, &|", ^A, Lev, RegOff(7), 0);
	RegLev(7):= Lev;
	PushFor(Lev, RegOff(7));
	];

TOS2Reg(7);			\r7:= control variable
DropV(1);
TOS2Stack;			\'for' loop limit
];	\GenFORSET



proc	GenFARG;		\$E6: Handle real argument
\This swaps the words in a real so that when GenARG pops them off the stack into
\ the heap they'll be in order. This may be a problem for external subroutines.
[Common($E6);
CleanReg(0);
FSP:= FSP-1;
Txt("	fmrrd	r14, r0, d%		@E6 FARG|", FSP);
Txt("	push	{r0, r14}|");
RegLev(0):= -1;  RegImm(0):= false;
DropV(1);			\drop the real
PushV(StkTyp, 0, 0);		\make it look like two integers for GenARG
PushV(StkTyp, 0, 0);
];	\GenFARG



proc	GenOPTBASE;		\$F7: Generate start of an optimized procedure
int	Lev;
[Lev:= HexIn(3);
Level:= Lev;
Common($F7);
Txt("	push	{r11, r14}|");
Txt("	ldr	r11, [r8, #base%-dseg]	@F7 OPTBASE|", Lev);
];	\GenOPTBASE



proc	GenWORD;		\$F9:
int	W, Ch;
[Common($F9);
Txt("	.word	");
loop	[if ChIn(3)=^L then ChOut(OD, ^L) else Backup;
	W:= IntIn(3);			\output decimal words until eol
	IntOut(OD, W);
	Backup;
	Ch:= ChIn(3);
	if Ch=CR ! Ch=LF then quit;
	ChOut(OD, ^,);
	];
CrLf(OD);
];	\GenWORD



proc	GenBYTE;		\$FA:
int	B, Ch;
[Common($FA);
Txt("	.byte	");
loop	[B:= HexIn(3);			\output hex bytes until eol
	Txt("0x");
	HexB(B);
	Backup;
	Ch:= ChIn(3);
	if Ch=CR ! Ch=LF then quit;
	ChOut(OD, ^,);
	];
CrLf(OD);
];	\GenBYTE



proc	GenABS;			\$56: Absolute value function
[Common($56);
if VSTyp(VSP-1)=ImmTyp then
	VSLev(VSP-1):= abs(VSLev(VSP-1))
else	[CleanReg(1);
	TOS2Reg(0);
	Txt("	cmp	r0, #0			@56 ABS|");
	Txt("	rsbmi	r0, #0|");
	RegLev(0):= -1;  RegVal(0):= abs(RegVal(0));
	DropV(1);
	PushV(RegTyp, 0, 0);
	];
];	\GenABS



proc	GenFABS;		\$57: Absolute value function of a real
[Common($57);
Txt("	fabsd	d%, d%  		@57 FABS|", FSP-1);
];	\GenFABS



proc	GenBASE;		\$F4: Generate the start of a procedure
\This sets up a base variable corresponding to the (static) nesting level
\ (base1, base2, etc). These base variables are used by any procedures nested
\ in the current procedure to access the current procedure's variables. To the
\ nested procedures the current procedure's variables will be "intermediate"
\ variables that are accessed via r12 -- they're not globals accessed via r10
\ nor locals accessed via r11. There is never a need to set up base0 because it
\ is never used.
\This also sets r11 to the current base level (> 0) to access local variables.
int	Lev, I;
[Lev:= HexIn(3);
Level:= Lev;
Common($F4);
OldLev:= -1;
for I:= 0 to RegMax do [RegLev(I):= -1;  RegImm(I):= false];
RegLev(7):= -1;  RegImm(7):= false;
RegLastUsed:= 0;
\\if Lev = 0 then return;

\The folowing isn't necessary when returning from level 0 because RET does a
\ simple return and base0 is never used:
\ Preserve r11 across procedure call.
\ Set base of local variables for current procedure.
\ Any nested procs might need base to access intermediate levels via r12.
\ Preserve baseLev across procedure call.

if NeedBase & 1<<Lev then
	[Txt("	ldr	r0, [r8, #base%-dseg]	@F4 BASE|", Lev);
	RegLev(0):= -1;  RegImm(0):= false;
	Txt("	push	{r0, r11, r12}|");		\save base, r11, and lr
	Txt("	ldr	r11, [r8, #baseary-dseg]|");	\set base for r12 access
	Txt("	str	r11, [r8, #base%-dseg]|", Lev);  
	]
else	[Txt("	push	{r11, r14}		@F4 BASE|");
	Txt("	ldr	r11, [r8, #baseary-dseg]|");
	];
];	\GenBASE



proc	GenDEND;		\$FD: End data segment; start code segment
[Common($FD);
Txt("	.text|	.align	2||");
];	\GenDEND



proc	GenDSTART;		\$FC: Start data segment
[Common($FC);
Txt("|	.data|	.align	2|");
];	\GenDSTART



proc	GenASCII;		\$FB: ASCII string
[Common($FB);
Txt("	.ascii	");
];	\GenASCII



proc	GenEND;			\$FE:
[Common($FE);
Txt("	.end|");		\This must match the size set up in natrpi.s
];	\GenEND



proc	GenStart;
[Txt("	.include	^"nat.inc^"|");
Txt("	.global	progrm|");
Txt("	.text|");
Txt("	.align	2|");
Txt("progrm:|");
];	\GenStart

\-------------------------------------------------------------------------------

proc	OpenFiles;	\Open input and output files specified on command line
int	InHand,		\input file handle
	OutHand,	\output file handle
	Sw, I, Ch;
char	CmdTail;

	func	GetSwitch;	\Find, remove, and return a command-line switch
	int	P, T;
	[for P:= 1 to CmdTail(0) do
	    if CmdTail(P) = ^- then
		[CmdTail(P):= $20;
		if P < CmdTail(0) then
			[T:= CmdTail(P+1);
			if T>=^A & T<=^Z then T:= T+$20;	\make lowercase
			CmdTail(P+1):= $20;
			return T;
			];
		];
	return 0;
	];	\GetSwitch


	proc	Parse;		\Parse command tail and set I/O handles
	char	ExtIn, ExtOut;
	int	P, ExtFlag, I;
	[ExtIn:= ".iil";   ExtOut:= ".s";
	P:= 1;
	ExtFlag:= false;
	loop	[if CmdTail(P) = ^. then  \parse command tail for extension
			[ExtFlag:= true;  quit];
		if CmdTail(P) = ^; then quit;	\ignore semicolon
		if P > CmdTail(0) then quit;	\if quit then P points to 0
		P:= P+1;
		];
	\Back up over any trailing spaces (caused by switches)
	repeat P:= P-1 until CmdTail(P) # ^ ;
	P:= P+1;
	if not ExtFlag then			\append input file extension
		for I:= 0 to 4 do CmdTail(P+I):= ExtIn(I);
	InHand:= FOpen(CmdTail+1, 0);
	for I:= 0 to 2 do CmdTail(P+I):= ExtOut(I);
	OutHand:= FOpen(CmdTail+1, 1);
	];	\Parse


[CmdTail:= Reserve($80+4);
I:= 1;				\get file name from command line
repeat	Ch:= ChIn(8);
	CmdTail(I):= Ch;
	I:= I+1;
until	Ch=0 ! Ch=EOF;
CmdTail(0):= I-2;	\number of characters in CmdTail not including last 0

loop	[Sw:= GetSwitch;
	case Sw of
	  ^a:	OD:= 0;
	   0:	quit			\no more switches on command line
	other	[Text(0, "UNRECOGNIZED SWITCH: -");  ChOut(0, Sw);  CrLf(0);
		Text(0, "Usage: iilo [option] filename
   -a: display Assembly code
");
		exit 1;
		];
	];
Parse;					\parse command line and set file handles
FSet(InHand, ^I);
FSet(OutHand, ^O);
];	\OpenFiles

\-------------------------------------------------------------------------------

proc	GenOp(Op);
int	Op;
[
cond false;
\debug
CrLf(OD);
Text(OD, " @@@@@@ Op = "); HexOut(OD, Op);
Text(OD, " RegLev(5) = "); IntOut(OD, RegLev(5));
Text(OD, " RegOff(5) = "); IntOut(OD, RegOff(5));
CrLf(OD);
\if ChIn(1) then [];
\end debug
cond true;


if Op < $48 then
	case Op of
	  $00:	GenEXIT;
	  $01:	GenLOD;
	  $02:	GenLDX;
	  $03:	GenSTO;
	  $04:	GenSTX;
	  $05:	GenCAL;
	  $06:	GenRET;
	  $07:	GenJMP;
	  $08:	GenJPC;
	  $09:	GenHPI;
	  $0A:	GenARG;
	  $0B:	GenIMM;
	  $0C:	GenCML;  
	  $0D:	GenADD;
	  $0E:	GenSUB;
	  $0F:	GenMUL;
	  $10:	GenDIV;
	  $11:	GenNEG;
	  $12, $13, $14, $15, $16, $17:
		GenCMP(Op);
	  $18:	GenFOR(false);
	  $19:	GenINP(false);
	  $1A:	GenOR;
	  $1B:	GenAND;
	  $1C:	GenNOT;
	  $1D:	GenEOR;
	  $1E:	GenDBA;
	  $1F:	GenSTD;
	  $20:	GenDBX;
	  $21:	GenADR;
	  $22:	GenFADR;
	  $23:	GenLDA;
	  $24:	GenIMM;
	  $25:	GenCJP;
	  $26:	GenJSR;
	  $27:	GenRTS;
	  $28:	GenDRP;
	  $29:	GenCEXT;
	  $2A:	GenFLOD;
	  $2B:	GenFSTO;
	  $2C:	GenFIMM;
	  $2D:	GenFADD;
	  $2E:	GenFSUB;
	  $2F:	GenFMUL;
	  $30:	GenFDIV;
	  $31:	GenFNEG;
	  $32, $33, $34, $35, $36, $37:
		GenFCMP(Op);
	  $38:	GenTRA;
	  $39:	GenTRX;
	  $3A:	GenTRI;
	  $3B:	GenSTT;
	  $3C:	GenMARK;
	  $3D:	GenASR;
	  $3E:	GenLSL;
	  $3F:	GenLSR
	other	[HexOut(0, Op);  CrLf(0);  Error(2)]
else	case Op of
	  $48:	GenARY;
	  $49:	GenFOR(true);
	  $4A:	GenINP(true);
	  $4B:	GenREM;
	  $4C:	GenSWAP;
	  $4D:	GenEXT;
	  $4E:	GenPIN;
	  $4F:	GenPOUT;
	  $50:	GenFIX;
	  $51:	GenFLOAT;
	  $52:	GenFSQRT;
	  $53:	GenSQRT;
	  $54:	GenFSQ;
	  $55:	GenSQ;
	  $56:	GenABS;
	  $57:	GenFABS;

	  $E5:	GenSTX2;
	  $E6:	GenFARG;
	  $E8:	GenFORSET;
	  $E9:	GenXR7;
	  $EB:	GenREG;
	  $EC:	GenSTK;
	  $ED:	GenCLEAN;
	  $EF:	GenVPSH;
	  $F0:	GenVPSHADR;
	  $F1:	GenVDRP;
	  $F2:	GenREAL;
	  $F4:	GenBASE;
	  $F5:	GenBASEARY;
	  $F6:	GenREGARY;
	  $F7:	GenOPTBASE;
	  $F9:	GenWORD;
	  $FA:	GenBYTE;
	  $FB:	GenASCII;
	  $FC:	GenDSTART;
	  $FD:	GenDEND;
	  $FE:	GenEND
	other	[HexOut(0, Op);  CrLf(0);  Error(2)];
];	\GenOp



int	I, Ch, Op;
[
VSTyp:= Reserve(VStkSize*IntSize);	\virtual stack
VSLev:= Reserve(VStkSize*IntSize);
VSOff:= Reserve(VStkSize*IntSize);
VSP:= 0;
ProcRets:= Reserve(LevelSize*IntSize);
for I:= 0 to LevelSize-1 do ProcRets(I):= 0;

Text(0, "
-- BACK-END ARM ASSEMBLY CODE GENERATOR FOR XPL0, VERSION 2.2 --

");
OD:= 3;
OpenFiles;
OpenO(OD);  OpenI(3);

Level:= 0;
ErrCnt:= 0;
LastOp:= -1;
NeedBase:= true;	\debug 0;
OldLev:= -1;
RegMax:= 6;
RegLastUsed:= 0;
for I:= 0 to 7 do
	[RegLev(I):= -1;
	 RegOff(I):= -1;
	 RegImm(I):= false;
	 RegVal(I):= 0;
	];
FSP:= 8;
ForSP:= 0;
GenStart;
LineCnt:= 1;
loop	[repeat Ch:= ChIn(3) until \Ch#LF &\ Ch#CR & Ch#$20 & Ch#$09;
	case Ch of
	  EOF:	quit;
	  LF:	LineCnt:= LineCnt+1;
	  ^L:	[ChOut(OD, Ch);		\data label
		repeat Ch:= ChIn(3);  ChOut(OD, Ch) until Ch=^:;
		];
	  ^K:	[			\"kode" label
		LastOp:= -1;	\Might not be coming from the previous opcode,
		OldLev:= -1;	\ so these registers could contain anything.
				\However the r7 register gets special handling,
				\ so don't bomb r7.
		for I:= 0 to RegMax do [RegLev(I):= -1;  RegImm(I):= false];
		RegLastUsed:= 0;
		ChOut(OD, Ch);
		repeat Ch:= ChIn(3);  ChOut(OD, Ch) until Ch=^:;
		];
	  ^":	[ChOut(OD, Ch);		\string
		repeat Ch:= ChIn(3);  ChOut(OD, Ch) until Ch = ^";
		CrLf(OD);		\(never includes any CRs or LFs)
		];
	  ^;:	[ChOut(OD, ^@);		\comment
		repeat Ch:= ChIn(3);  ChOut(OD, Ch) until Ch = LF;
		CrLf(OD);
		LineCnt:= LineCnt+1;
		];
	  ^`:	[repeat Ch:= ChIn(3);  ChOut(OD, Ch) until Ch = LF;    \asm code
		LineCnt:= LineCnt+1;
		for I:= 0 to 3 do	\bomb registers that were possibly used
			[RegLev(I):= -1;  RegImm(I):= false];
		LastOp:= -1;
		OldLev:= -1;		\system calls can bomb r12
		]
	other	[Backup;
		Op:= HexIn(3);
		if Op>=$12 & Op<=$17 \CMP\ then LastOp:= Op
		else if LastOp>=$12 & LastOp<=$17 \CMP\ then 
			[if Op = $08\JPC\ then GenJOC(LastOp\CMP\)
			else	[GenOp(LastOp);  GenOp(Op)];
			]
		else	GenOp(Op);
		];
	];
GenEND;
Close(OD);

CrLf(0);
Text(0, "Errors detected: ");  IntOut(0, ErrCnt);  CrLf(0);
CrLf(0);

return if ErrCnt#0 then 1 else 0;
];	\Main
